from typing import List

a = 5
b = ""
class A:
    def __init__(self):
        self._something = "Hello"
        self._someOtherThing = "HHH"
        self._my_dict = {}
        self._private = []
        self.a = A()
        self._private.append(4)
        self.a._private.append(5)

    def method(self):
        """Some docstrings"""
        pass


class B(A):
    def method(self):
        global a
        pass

    def method2(self, arg1: str, arg2, arg3, arg4, arg5):
        """
        short desc
        Parameters:
            arg1: Some arg
            arg2 (str): Another arg
            arg3: Another
            arg4: Ohter
        """

def func(test: str, argumentCamel, A: List[int]):
    someVariable = "A"
    Another = "B"
    _HeHe = "C"
    for badName in range(3):
        print()
    global b

def camelCase(test_):
    pass

a = 1
X = 2
some_list = []
my_dict = {}
my_dictionary = {"a": 3}
instance = A()
A.method(instance)  # instance.method()
instance._private.append(3)
instance._private = []
instance.a._private.append(1)
if True:
    with open("a"):
        while True:
            for i in range(10):
                try:
                    if True:
                        pass
                except:
                    pass
                finally:
                    pass

elif a == True:
    pass
elif True:
    pass
elif True:
    pass
elif True:
    pass
elif True:
    pass
elif True:
    pass
elif True:
    pass
elif True:
    pass
else: pass


for x in range(10):
    print(x)

for i in range(10):
    print(i)

if True:
    pass
elif True:
    pass

if True:
    pass
else:
    if True:
        pass

hungarian_str = ""
integration = 1
middle_int_hey = 1
dictator = 1
mint = 2



"""
Analyses code structure
"""
from astroid.mixins import BlockRangeMixIn
from astroid.node_classes import NodeNG, Compare, Const
from astroid import Pass, TryExcept

from .analyser import Analyser, register_check


class StructureAnalyser(Analyser):
    """Structure analyser, checks if general structure is ok"""

    @register_check("{3}: Number of lines longer than {0} characters: {1}\n"
                    "\tLines {2} are longer than {0} characters")
    def check_line_length(self):
        """Checks if code has any line that's too long """
        # character_count, number_of_lines, line_numbers
        result: list[tuple[int, int, str, str]] = []
        for filename, attr in self._sources.items():
            for length in [80, 100, 120, 140]:
                long_lines = [str(line_num + 1)
                              for line_num, line in enumerate(attr.source)
                              if len(line.rstrip()) > length]
                if not long_lines:
                    continue
                result.append((length, len(long_lines),
                               ", ".join(long_lines) or None,
                               filename))
        return result

    @register_check("{}:{}-{}: too many nested control structures\n{}")
    def check_structure_complexity(self):
        """Checks if control structures are nested too deeply"""
        # line_start, line_end, block
        result: list[tuple[str, int, int, str]] = []
        for filename, attr in self._sources.items():
            complexities: dict[NodeNG, int] = {}
            for node in attr.tree.post_order():
                if not isinstance(node, BlockRangeMixIn):
                    continue
                children = node.body.copy()
                if isinstance(node, TryExcept):
                    children.extend(node.handlers)
                if hasattr(node, "orelse") and len(node.orelse) > 1:
                    children.extend(child for child in node.orelse)
                complexities[node] = 1 + max((complexities.get(child, 0)
                                              for child in node.body),
                                             default=0)
            checked_nodes = []
            for node, complexity in complexities.items():
                if complexity < 4:
                    continue
                highest_node = node
                while complexities.get(highest_node.parent) is not None:
                    highest_node = highest_node.parent
                if highest_node in checked_nodes:
                    continue
                start, end = highest_node.block_range(highest_node.lineno)
                result.append((filename, start, end,
                               "\n".join(attr.source[start - 1:end])))
                checked_nodes.append(highest_node)
        return result

    @register_check("{}:{}: Control structure block does nothing\n")
    def check_structure_empty(self):
        """Checks if control structures blocks just have `pass` in them"""
        # line_start, line_end, block
        result: list[tuple[str, int]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.post_order():
                if not isinstance(node, BlockRangeMixIn):
                    continue
                if len(node.body) == 1 and \
                        isinstance((pass_ := node.body[0]), Pass):
                    result.append((filename, pass_.lineno))
                elif getattr(node, "orelse", None) is not None and \
                        len(node.orelse) == 1 and \
                        isinstance((pass_ := node.orelse[0]), Pass):
                    result.append((filename, pass_.lineno))
        return result

    @register_check("{}:{}: Redundant boolean (in)equality\n{}")
    def check_redundant_boolean_equality(self):
        """ Check for condition == True or condition == False """
        # filename, lineno, line
        result: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.post_order():
                if not isinstance(node, Compare):
                    continue
                for operator, right_side_node in node.ops:
                    if operator != "==" and operator != "!=":
                        continue
                    if not isinstance(right_side_node, Const):
                        continue
                    right_side_value = right_side_node.value
                    if not isinstance(right_side_value, bool):
                        continue
                    result.append((filename, node.lineno,
                                   attr.source[node.lineno - 1]))
        return result



"""
Anything violation related
"""
from dataclasses import dataclass
from typing import Union


ViolationResult = tuple[Union[str, int, float]]


@dataclass
class BaseViolation:
    """Represents a list of violations found after running the checkers"""
    description: str
    format: str
    values: list[ViolationResult]
    



"""
Abstract formatter
Used to convert error outputs to user friendly output to be displayed on user
interface
"""
from ..type_hints import Config, CheckOutput


class Formatter:
    """Abstract formatter"""
    def __init__(self, config: Config, check_output: CheckOutput):
        """
        Constructor
        Args:
            config: configuration file that links criteria to tests
            check_output: output from checkers
        """
        self._config = config
        self._check_output = check_output

    def format(self) -> str:
        """Returns formatted string to be displayed on screen"""
        raise NotImplementedError

    def get_config(self) -> Config:
        """Returns configuration object"""
        return self._config

    def get_check_output(self):
        """Returns check output from analysers"""
        return self._check_output



"""
String helper functions
"""
import textwrap


def indent(multiline_str: str, indented=4):
    """
    Converts a multiline string to an indented string

    Args:
        multiline_str: string to be converted
        indented: number of space used for indentation

    Returns: Indented string
    """
    return textwrap.indent(multiline_str, " " * indented)


def new_line(string: str):
    """
    Append a new line at the end of the string

    Args:
        string: String to make a new line on

    Returns: Same string with a new line character
    """
    return string + "\n"



"""
Analyse encapsulation violations
"""
from astroid import Attribute, Name, AssignAttr

from .analyser import Analyser, register_check


class EncapsulationAnalyser(Analyser):
    """
    Encapsulation analyser, checks if class members follow best
    access control practices
    """
    @register_check("{}:{}: Private attribute `{}` accessed outside class:\n"
                    "\t{}")
    def check_private_attribute_accessed_outside_class(self):
        """Checks if private attribute is accessed outside class """
        result: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, Attribute):
                    continue
                if not node.attrname.startswith("_") or \
                        node.attrname.startswith("__"):
                    continue
                expr = node.expr
                if isinstance(expr, Name) and expr.name == "self":
                    continue
                result.append((filename, node.lineno, node.attrname,
                               self.get_line(filename, node.lineno)))
        return result

    @register_check("{}:{}: Private attribute `{}` defined outside class:\n"
                    "\t{}")
    def check_private_attribute_defined_outside_class(self):
        """Checks if private attribute is defined outside class """
        result: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignAttr):
                    continue
                if not node.attrname.startswith("_") or \
                        node.attrname.startswith("__"):
                    continue
                expr = node.expr
                if isinstance(expr, Name) and expr.name == "self":
                    continue
                result.append((filename, node.lineno, node.attrname,
                               self.get_line(filename, node.lineno)))
        return result

    @register_check("{}:{}: Public attribute `{}` could have been private:\n"
                    "\t{}")
    def check_public_attribute_defined_self(self):
        """Checks if public attribute could be replaced by a private attribute """
        result: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignAttr):
                    continue
                if node.attrname.startswith("_") and \
                        not node.attrname.startswith("__"):
                    continue
                expr = node.expr
                if not (isinstance(expr, Name) and expr.name == "self"):
                    continue
                result.append((filename, node.lineno, node.attrname,
                               self.get_line(filename, node.lineno)))
        return result



"""
Type hints helpers
"""
from dataclasses import dataclass

from .utils.tree import SyntaxTree
from .utils.violation import BaseViolation


CriteriaConfig = dict[str, list]

Config = dict[str, CriteriaConfig]

AnalyserResults = dict[str, BaseViolation]

CheckOutput = dict[str, AnalyserResults]

LintResults = dict[str, dict[str, BaseViolation]]


@dataclass
class AnalyserHelper:
    tree: SyntaxTree
    source: list[str]


"""
Tree data structure helper to traverse through syntax nodes
"""

from typing import Iterator
from astroid.node_classes import NodeNG


class SyntaxTree:
    """Data structure helper to traverse syntax nodes"""
    def __init__(self, root: NodeNG):
        """
        Constructor
        Parameters:
            root: root node of this tree
        """
        self._root = root

    def get_root(self) -> NodeNG:
        """Returns the root node"""
        return self._root

    @staticmethod
    def post_order_traversal(node: NodeNG) -> Iterator[NodeNG]:
        """
        Yields nodes in post order, starting from `node`
        Parameters:
            node: Node to start yielding form
        """
        for child_node in node.get_children():
            yield from SyntaxTree.post_order_traversal(child_node)
        yield node

    @staticmethod
    def pre_order_traversal(node: NodeNG) -> Iterator[NodeNG]:
        """
        Yields nodes in pre order, starting from `node`
        Parameters:
            node: Node to start yielding form
        """
        yield node
        for child_node in node.get_children():
            yield from SyntaxTree.pre_order_traversal(child_node)

    def post_order(self) -> Iterator[NodeNG]:
        """Yields node in post order, starting from root"""
        yield from self.post_order_traversal(self._root)

    def pre_order(self) -> Iterator[NodeNG]:
        """Yields node in pre order, starting from root"""
        yield from self.pre_order_traversal(self._root)



"""
Analyser for class vs instance usage
"""
from astroid import ClassDef, Call, Attribute, Name
from .analyser import Analyser, register_check


class ClassInstanceAnalyser(Analyser):
    """Checks violations in class vs instance usages"""
    @register_check("{}:{}: Method call treated as function instead of "
                    "member of instance:\n\t{}")
    def check_method_call(self):
        """Checks if method call is treated as regular function instead of
        member of class instance"""
        # filename, lineno, line
        result: list[tuple[str, int, str]] = []
        class_names: list[str] = []
        for filename, attr in self._sources.items():
            tree = attr.tree
            for node in tree.pre_order():
                if not isinstance(node, ClassDef):
                    continue
                class_names.append(node.name)

            for node in tree.pre_order():
                if not isinstance(node, Call):
                    continue
                func = node.func
                if not isinstance(func, Attribute):
                    continue
                if not isinstance(func.expr, Name):
                    continue
                if func.expr.name in class_names:
                    result.append((filename, node.lineno, node.as_string()))
        return result



import sys
from pathlib import Path
from typing import Type

import yaml
from astroid import parse, Module

from .analysers import Analyser
from .formatters import Formatter
from .type_hints import AnalyserHelper, LintResults
from .utils.tree import SyntaxTree


class Run:
    def __init__(self, analysers: list[Type[Analyser]],
                 formatter: Type[Formatter],
                 source_file_names: list[str], config_file_name: str):
        self._analysers = analysers
        self._formatter_cls = formatter
        self._analyser_helpers: dict[str, AnalyserHelper] = {}
        for source_file_name in source_file_names:
            try:
                with open(source_file_name) as fin:
                    source = fin.read()
            except FileNotFoundError:
                print(f"{source_file_name} not found, skipping...",
                      file=sys.stderr)
                continue
            module: Module = parse(source)
            lines = source.splitlines()
            tree = SyntaxTree(module)
            filename = Path(source_file_name).name
            self._analyser_helpers[filename] = AnalyserHelper(tree, lines)

        with open(config_file_name) as config_file:
            self._config = yaml.load(config_file, Loader=yaml.SafeLoader)
        self._results: LintResults = {}
        # Uncomment to see source tree
        # print(module.repr_tree())

    def run(self):
        for analyser_cls in self._analysers:
            analyser = analyser_cls(self._analyser_helpers)
            analyser.run()
            self._results[analyser_cls.__name__] = analyser.get_results()

    def print_results(self):
        formatter = self._formatter_cls(self._config, self._results)
        print(formatter.format())

    def get_results(self):
        return self._results



"""__init__"""
from .analyser import Analyser
from .class_instance import ClassInstanceAnalyser
from .encapsulation import EncapsulationAnalyser
from .naming import NamingAnalyser
from .scope import ScopeAnalyser
from .structure import StructureAnalyser



import dataclasses
import json


class DataclassJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



"""
Analyse scope violations
"""
from astroid import Global, Const

from .analyser import Analyser, register_check


class ScopeAnalyser(Analyser):
    """Analyser checking for scope violations"""
    MAGIC_WHITELIST = [0, 1, -1, 100]

    @register_check("{}:{}: Globals used:\n\t{}")
    def check_globals(self):
        """Checks if code has any global variables"""
        # lineno, line
        result: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, Global):
                    continue
                result.append((filename, node.lineno,
                               self.get_line(filename, node.lineno)))
        return result

    @register_check("{}:{}: Magic number used\n\t{}")
    def check_magic_numbers_used(self):
        """Check if any magic number has been used """
        # lineno, line
        result: list[tuple[str, int, str]] = []
        checked_lines = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, Const):
                    continue
                if node.value in self.MAGIC_WHITELIST:
                    continue
                if not isinstance(node.value, int) and \
                        not isinstance(node.value, float):
                    continue
                if node.lineno in checked_lines:
                    continue
                result.append((filename, node.lineno,
                               self.get_line(filename, node.lineno)))
                checked_lines.append(node.lineno)
        return result



"""
Abstract analyser
"""
from functools import wraps
from inspect import getmembers, ismethod
from typing import Callable

from ..type_hints import AnalyserResults, AnalyserHelper
from ..utils import SyntaxTree, BaseViolation, ViolationResult


def register_check(error_format: str):
    """
    Registers a new checker to an analyser
    Args:
        error_format: error format of violation
    """
    def decorator(check_method: Callable):
        @wraps(check_method)
        def wrapper(*args, **kwargs):
            analyser = args[0]
            checker_name = check_method.__name__
            analyser.register_checker(checker_name,
                                      check_method.__doc__,
                                      error_format)
            result: list[ViolationResult] = check_method(*args, **kwargs)
            analyser.add_violations(checker_name, result)
        return wrapper
    return decorator


class Analyser:
    """Abstract base analyser"""
    def __init__(self, sources: dict[str, AnalyserHelper]):
        """
        Constructor
        Args:
            tree: syntax tree
            source: list of lines from source code
        """
        self._check_results: AnalyserResults = {}
        self._sources = sources

    def register_checker(self, name: str, description: str, error_format: str):
        """
        Registers a new checker to this analyser
        Args:
            name: name of the checker, typically the method name
            description: description of this checker
            error_format: format string used to display violations
        """
        self._check_results[name] = BaseViolation(description, error_format, [])

    def get_results(self) -> AnalyserResults:
        """
        Returns results of all checkers of this analyser
        """
        return self._check_results

    def add_violations(self, checker_name: str,
                       results: list[ViolationResult]) -> None:
        """
        Adds violation results to a checker
        Args:
            checker_name: name of the checker
            results: list of violation results
        """
        self._check_results[checker_name].values.extend(results)

    def get_line(self, file_name: str, line_number: int) -> str:
        """Returns line given line number"""
        return self._sources[file_name].source[line_number - 1].strip()

    def run(self):
        """
        Runs all checkers
        """
        for method_name, method in getmembers(self, predicate=ismethod):
            if not method_name.startswith("check_"):
                continue
            method()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="mikelint",
    version="1.2.2",
    description="Linter used for CSSE1001 at UQ",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/mike-fam/mikelint",
    author="Mike Pham",
    author_email="mikepham1207@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["mikelint"],
    include_package_data=True,
    install_requires=[
        "astroid>=2.5",
        "isort>=5.8",
        "lazy-object-proxy>=1.6",
        "mccabe>=0.6",
        "PyYAML>=5.4",
        "toml>=0.10",
        "wrapt>=1.12"
    ],
    python_requires=">=3.8",
    # package_dir={"": ""}
)



"""Naming analyser, checks everything name related"""
import re
from astroid import AssignName, AssignAttr, For, Call, FunctionDef

from .analyser import Analyser, register_check


class NamingAnalyser(Analyser):
    """Analyse good naming"""
    SNAKE_CASE = re.compile(r"[a-z_][a-z0-9_]{0,30}$")
    CONSTANT_SNAKE_CASE = re.compile(r"(([A-Z_][A-Z0-9_]*)|(__.*__))$")
    HUNGARIAN_NOTATION = re.compile(r"(?<![^\W_\d])"
                                    r"(str|string|int|list|lst|dict|dictionary|"
                                    r"tup|tuple|float|func|function|method|chr|"
                                    r"character)"
                                    r"(?![^\W_\d])")

    @register_check(error_format="{}:{}: Non snake-case variable '{}':\n\t{}")
    def check_variable_snake_case(self):
        """Checks for non snake-case variable naming """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignName):
                    continue

                # Ignore constants
                if node.parent is attr.tree.get_root():
                    continue

                if self.SNAKE_CASE.fullmatch(node.name):
                    continue
                results.append((filename, node.lineno, node.name,
                                node.parent.as_string().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Non snake-case function/method name "
                                 "'{}':\n\t{}")
    def check_method_snake_case(self):
        """Checks for non snake-case function and method naming """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                if self.SNAKE_CASE.fullmatch(node.name):
                    continue
                results.append((filename, node.lineno, node.name,
                                node.as_string().strip().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Non snake-case attribute '{}':\n\t{}")
    def check_attribute_snake_case(self):
        """Checks for non snake-case attribute naming """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignAttr):
                    continue
                if self.SNAKE_CASE.fullmatch(node.attrname):
                    continue
                results.append((filename, node.lineno, node.attrname,
                                node.parent.as_string().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Potential bad variable name '{}':\n"
                                 "\t{}")
    def check_potential_bad_variable_names(self):
        """Checks for potential bad naming, i.e. names with 1-2 characters """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignName):
                    continue

                # skip naming for `for i in range`
                if self._is_for_range_variable(node) \
                        and node.name in ["i", "j", "k"]:
                    continue

                if len(node.name) >= 3:
                    continue

                results.append((filename, node.lineno, node.name,
                                node.parent.as_string().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Hungarian notation variable '{}':\n"
                                 "\t{}")
    def check_hungarian_notation_variable(self):
        """Checks for variable names with hungarian notation """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignName):
                    continue
                if not self.HUNGARIAN_NOTATION.search(node.name):
                    continue
                results.append((filename, node.lineno, node.name,
                                node.parent.as_string().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Hungarian notation attribute '{}':\n"
                                 "\t{}")
    def check_hungarian_notation_attribute(self):
        """Checks for attribute names with hungarian notation """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, AssignAttr):
                    continue
                if not self.HUNGARIAN_NOTATION.search(node.attrname):
                    continue
                results.append((filename, node.lineno, node.attrname,
                                node.parent.as_string().splitlines()[0]))
        return results

    @register_check(error_format="{}:{}: Hungarian notation in function/method "
                                 "name '{}':\n\t{}")
    def check_hungarian_notation_method(self):
        """Checks for hungarian notation function and method naming """
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                if not self.HUNGARIAN_NOTATION.search(node.name):
                    continue
                results.append((filename, node.lineno, node.name,
                                node.as_string().strip().splitlines()[0]))
        return results

    @register_check("{}:{}: Variables defined at global scope should be "
                    "treated as constants:\n\t{}")
    def check_constant_naming(self):
        """Checks if constants are named in UPPER_SNAKE_CASE"""
        # lineno, line
        results: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.get_root().get_children():
                if not isinstance(node, AssignName):
                    continue
                if not self.CONSTANT_SNAKE_CASE.fullmatch(node.name):
                    continue
                results.append((filename, node.lineno, node.as_string()))
        return results

    @staticmethod
    def _is_for_range_variable(node: AssignName):
        """Check if the assignment is for a variable for a for range loop """
        parent = node.parent
        if not isinstance(parent, For):
            return False
        iter_on = parent.iter
        if not isinstance(iter_on, Call):
            return False
        return iter_on.func.as_string() == "range"



"""Docstring analyser"""
from typing import Optional
from astroid import ClassDef, FunctionDef
from docstring_parser import parse, Style

from .analyser import Analyser, register_check


class DocstringAnalyser(Analyser):
    """Analyses anything docstring related"""
    @register_check("{}:{}: Class {} missing docstring")
    def check_class_docstrings(self):
        """Checks if classes missing docstrings """
        result: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, ClassDef):
                    continue
                if node.doc:
                    continue
                result.append((filename, node.lineno, node.name,))
        return result

    @register_check("{}:{}: Method/Function {} missing docstring")
    def check_methods_function_docstrings(self):
        """Checks if methods or functions missing docstrings """
        result: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                if node.doc:
                    continue
                result.append((filename, node.lineno,
                               self._get_method_name(node)))
        return result

    @register_check("{}:{}: Missing/extra fields in docstrings of "
                    "function/method:\n"
                    "\tFunction/Method name: {}\n"
                    "\tMissing fields: {}\n"
                    "\tExtra fields: {}\n")
    def check_method_docstring_unexpected_missing(self):
        """Checks if docstring has missing or extra fields """
        result: list[tuple[str, int, str, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                signatures = self._get_expected_and_actual_method_parameters(node)
                if signatures is None:
                    continue
                method_name = self._get_method_name(node)
                expected, actual = signatures
                unexpected = [arg for arg, _ in actual]
                missing = []
                for arg, _ in expected:
                    try:
                        unexpected.remove(arg)
                    except ValueError:
                        missing.append(arg)
                if not unexpected and not missing:
                    continue
                result.append((filename, node.lineno, method_name,
                               ", ".join(missing) or "None",
                               ", ".join(unexpected) or "None"))
        return result

    @staticmethod
    def _get_expected_and_actual_method_parameters(node: FunctionDef) -> \
            (list[tuple[str, Optional[str]]], list[tuple[str, Optional[str]]]):
        if not node.doc:
            return None
        try:
            docstring = parse(node.doc, Style.google)
        except:
            return None
        actual = [(param.arg_name, param.type_name)
                  for param in docstring.params]
        expected = [(arg.name, annotation.as_string())
                    for arg, annotation in zip(node.args.args,
                                               node.args.annotations)
                    if annotation is not None]
        # Remove 'self' from expected signature
        if isinstance(node.parent, ClassDef):
            try:
                expected.pop(0)
            except IndexError:
                # for problems such as
                # class A:
                #     def method():  # no self here
                # idek why they do this, but they did
                pass

        return expected, actual

    @staticmethod
    def _get_method_name(node: FunctionDef):
        method_name = node.name
        parent = node.parent
        if isinstance(parent, ClassDef):
            method_name = f"{parent.name}.{method_name}"
        return method_name

    @register_check("{}:{}: Missing parameter types in docstrings of "
                    "function/method:"
                    "\n\tFunction/Method name: {}\n"
                    "\tFields that lack param type: {}\n")
    def check_method_docstring_missing_type(self):
        """Checks if docstrings have all the parameter types necessary"""
        # method_name, lineno, fields
        results: list[tuple[str, int, str, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                signatures = self._get_expected_and_actual_method_parameters(node)
                if signatures is None:
                    continue
                method_name = self._get_method_name(node)
                expected, actual = signatures
                expected = dict(expected)
                actual = dict(actual)
                fields = []
                for name, type_ in actual.items():
                    if name not in expected:
                        continue
                    if type_ is None and expected[name] is None:
                        fields.append(name)
                if fields:
                    results.append((filename, node.lineno, method_name,
                                    ", ".join(fields)))
        return results

    @register_check("{}:{}: Cannot parse docstring of function/method:\n"
                    "\t{}")
    def check_docstring_correct_format(self):
        """Checks if docstring can be parsed correctly """
        result: list[tuple[str, int, str]] = []
        for filename, attr in self._sources.items():
            for node in attr.tree.pre_order():
                if not isinstance(node, FunctionDef):
                    continue
                if not node.doc:
                    continue
                method_name = self._get_method_name(node)
                try:
                    parse(node.doc, Style.google)
                except:
                    result.append((filename, node.lineno, method_name))
        return result



"""
A simple formatter to display analyser output in similar format as testrunner
"""
from .formatter_ import Formatter
from ..utils import new_line, indent, BaseViolation

BLOCK_WIDTH = 80
HORIZONTAL_BORDER = "-" * (BLOCK_WIDTH - 2)
BLOCK_TEMPLATE = """\
/{0}\\
|{{:^{1}}}|
\\{0}/\
""".format(HORIZONTAL_BORDER, BLOCK_WIDTH - 2)


class SimpleFormatter(Formatter):
    """Simple formatter with output similar to testrunner"""
    def format(self) -> str:
        output = ""
        for criteria_name, criteria in self._config.items():
            output += new_line(BLOCK_TEMPLATE.format(criteria_name))
            for index, (sub_criteria, analysers) in \
                    enumerate(criteria.items(), start=1):
                output += new_line(indent(
                    self._format_criteria(index, sub_criteria, analysers)))
                if index < len(criteria):
                    output += new_line(HORIZONTAL_BORDER)
        output += HORIZONTAL_BORDER
        return output

    def _format_criteria(self, index: int, sub_criteria: str,
                         checkers: list[str]):
        """
        Returns a string containing useful information from the specified
        analyser

        Args:
            index: index of the sub criteria
            sub_criteria: name of sub criteria
            checkers: list of string representations of checkers, might be in
                the format of Analyser.checker or Analyser

        Returns: Meaningful string containing information of the checker(s).
        """
        output = new_line(f"{index}. {sub_criteria}")
        for analyser in checkers:
            analyser_name, _, checker_name = analyser.partition(".")
            if checker_name:
                output += new_line(indent(self.format_violation(
                    self._check_output[analyser_name][checker_name])))
            else:
                for violation in self._check_output[analyser_name].values():
                    output += new_line(indent(self.format_violation(violation)))
        if not checkers:
            output += new_line("None")
        return output

    @staticmethod
    def format_violation(violation: BaseViolation):
        """
        Gets a meaningful string containing information about a violation

        Args:
            violation: BaseViolation information

        Returns: a meaningful string containing information about a violation
        """
        output = new_line(violation.description)
        for format_values in violation.values:
            output += new_line(indent(violation.format.format(*format_values)))
        if not violation.values:
            output += new_line(indent("No violations found"))
        return output



import json

from ..utils import BaseViolation, DataclassJsonEncoder
from .formatter_ import Formatter


class JsonFormatter(Formatter):
    def format(self) -> str:
        output = {}
        for criteria_name, criteria in self._config.items():
            output[criteria_name] = criteria_result = {}
            for sub_criteria, analysers in criteria.items():
                criteria_result[sub_criteria] = self._format_criteria(analysers)

        return json.dumps(output, indent=4, cls=DataclassJsonEncoder)

    def _format_criteria(self, checkers: list[str]) -> list[dict]:
        """
        Returns a string containing useful information from the specified
        analyser

        Args:
            checkers: list of string representations of checkers, might be in
                the format of Analyser.checker or Analyser

        Returns: Meaningful string containing information of the checker(s).
        """
        output = []
        for analyser in checkers:
            analyser_name, _, checker_name = analyser.partition(".")
            if checker_name:
                output.append(self.serialise_violation(
                    self._check_output[analyser_name][checker_name]))
            else:
                for violation in self._check_output[analyser_name].values():
                    output.append(self.serialise_violation(violation))
        return output

    @staticmethod
    def serialise_violation(violation: BaseViolation):
        """
        Gets a meaningful string containing information about a violation

        Args:
            violation: BaseViolation information

        Returns: a meaningful string containing information about a violation
        """
        serialised_violation = {"description": violation.description,
                                "violations": []}
        for value in violation.values:
            serialised_violation["violations"].append(
                violation.format.format(*value))
        return serialised_violation



import argparse

from .formatters import SimpleFormatter, JsonFormatter
from .analysers import (
    ClassInstanceAnalyser,
    EncapsulationAnalyser, NamingAnalyser,
    ScopeAnalyser, StructureAnalyser
)
from .run import Run


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", help="Configuration file",
                        required=True)
    parser.add_argument("-s", "--source", help="Source file",
                        required=True, action="append")
    parser.add_argument("-j", "--json", help="Display JSON output",
                        action="store_true", default=False)
    args = parser.parse_args()
    analysers = [
        ClassInstanceAnalyser,
        EncapsulationAnalyser,
        NamingAnalyser,
        ScopeAnalyser,
        StructureAnalyser
    ]
    runner = Run(analysers,
                 JsonFormatter if args.json else SimpleFormatter,
                 args.source,
                 args.config)
    runner.run()
    runner.print_results()


if __name__ == '__main__':
    main()
